package at.ac.uibk.gitsearch.repository.search;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.es.model.DocumentInfo;
import at.ac.uibk.gitsearch.service.dto.GitFilesAggregationDTO;
import at.ac.uibk.gitsearch.service.dto.GitFilesDTO;
import at.ac.uibk.gitsearch.service.dto.GitFilesPageDetailsDTO;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.aggregations.AggregationBuilders;
import org.elasticsearch.search.aggregations.bucket.terms.TermsAggregationBuilder;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.stereotype.Repository;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

@Repository
public class GitFilesRepositoryImpl implements GitFilesRepository {

    private final RestHighLevelClient elasticsearchClient;

    private final ApplicationProperties properties;

    private final int FRAGMENT_SIZE = 200;

    public GitFilesRepositoryImpl(RestHighLevelClient elasticsearchClient, ApplicationProperties properties) {
        this.elasticsearchClient = elasticsearchClient;
        this.properties = properties;
    }

    @Override
    public GitFilesPageDetailsDTO pageDetails(SearchInputDTO searchInputDTO, int pageSize) throws IOException {
        Collection<String> projectIDs = searchProjectsIDsMetadata(searchInputDTO);
        SearchRequest searchRequest = new SearchRequest(GitFilesRepositoryConstants.INDEX_FULLTEXT);

        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery()
            .must(QueryBuilders
                .simpleQueryStringQuery(searchInputDTO.getFulltextQuery())
                .field(GitFilesRepositoryConstants.FULLTEXT_CONTENT));


        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedRepository(), GitFilesRepositoryConstants.FULLTEXT_REPOSITORY);
        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedRepository(), GitFilesRepositoryConstants.FULLTEXT_SUB_GROUP);
        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedFileFormat(), GitFilesRepositoryConstants.FULLTEXT_FILE_EXTENSION);
        if (projectIDs != null) {
            queryBuilder
                .filter(QueryBuilders.termsQuery(GitFilesRepositoryConstants.FULLTEXT_PROJECT_ID, projectIDs));
        }

        char[] boundaryChars = {'\n'};

        HighlightBuilder highlightBuilder = new HighlightBuilder()
            .highlighterType("plain")
            .preTags(properties.getSearch().getHighlightPre())
            .postTags(properties.getSearch().getHighlightPost())
            .field(GitFilesRepositoryConstants.FULLTEXT_CONTENT)
            .boundaryChars(boundaryChars)
            .boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE)
            .fragmentSize(FRAGMENT_SIZE);

        //String[] includes = {};
        //String[] excludes = {HighlightRepositoryConstants.CONTENT_FIELD};

        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
        sourceBuilder.query(queryBuilder)
            .highlighter(highlightBuilder)
            .size(pageSize)
            .from((searchInputDTO.getPage() - 1) * pageSize);
        //.fetchSource(includes, excludes);

        searchRequest.source(sourceBuilder);

        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        List<GitFilesDTO> gitFilesDTOS = new ArrayList<>();
        for (SearchHit searchHit : searchResponse.getHits().getHits()) {
            DocumentInfo documentInfo = objectMapper.readValue(searchHit.getSourceAsString(), DocumentInfo.class);
            gitFilesDTOS.add(new GitFilesDTO(documentInfo, searchHit, properties));
        }
        long hitCount = searchResponse.getHits().getTotalHits();
        return new GitFilesPageDetailsDTO(gitFilesDTOS, hitCount);
    }

    private static void addFulltextSelection(BoolQueryBuilder queryBuilder, List<String> selectedFileFormat, String fileExtensionField) {
        if (selectedFileFormat.isEmpty()) {
            return;
        }
        queryBuilder.must(QueryBuilders.termsQuery(fileExtensionField, selectedFileFormat));
    }

    @Override
    public GitFilesAggregationDTO aggregation(String query) throws IOException {
        SearchRequest searchRequest = new SearchRequest(GitFilesRepositoryConstants.INDEX_FULLTEXT);

        BoolQueryBuilder matchQueryBuilder = QueryBuilders.boolQuery()
            .must(QueryBuilders.simpleQueryStringQuery(query).field(GitFilesRepositoryConstants.FULLTEXT_CONTENT));

        TermsAggregationBuilder groupSubGroup = AggregationBuilders
            .terms(GitFilesRepositoryConstants.FULLTEXT_SUB_GROUP_AGG)
            .field(GitFilesRepositoryConstants.FULLTEXT_SUB_GROUP);

        TermsAggregationBuilder groupFileExtension = AggregationBuilders
            .terms(GitFilesRepositoryConstants.FULLTEXT_FILE_EXTENSION_AGG)
            .field(GitFilesRepositoryConstants.FULLTEXT_FILE_EXTENSION);

        TermsAggregationBuilder groupRepository = AggregationBuilders
            .terms(GitFilesRepositoryConstants.FULLTEXT_REPOSITORY_AGG)
            .field(GitFilesRepositoryConstants.FULLTEXT_REPOSITORY);

        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder()
            .query(matchQueryBuilder)
            .size(0)
            .aggregation(groupSubGroup)
            .aggregation(groupFileExtension)
            .aggregation(groupRepository);

        searchRequest.source(sourceBuilder);

        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

        return new GitFilesAggregationDTO(searchResponse.getAggregations());
    }

    private Collection<String> searchProjectsIDsMetadata(SearchInputDTO searchInputDTO) throws IOException {
        if (!searchInputDTO.hasMetadataInput()) {
            return null;
        }
        SearchRequest searchRequest = new SearchRequest(GitFilesRepositoryConstants.INDEX_METADATA);
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
        addMetadataQueryShould(queryBuilder, searchInputDTO.getMetadataProgrammingLanguage(), GitFilesRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES);
        addMetadataQueryShould(queryBuilder, searchInputDTO.getMetadataKeywords(), GitFilesRepositoryConstants.METADATA_KEYWORDS);
        addMetadataQueryMust(queryBuilder, searchInputDTO.getMetadataAuthor(), GitFilesRepositoryConstants.METADATA_CREATOR);
        addMetadataQueryMust(queryBuilder, searchInputDTO.getMetadataLicense(), GitFilesRepositoryConstants.METADATA_LICENSE);
        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder()
            .query(queryBuilder)
            .fetchSource(GitFilesRepositoryConstants.METADATA_PROJECT_ID, null);
        // TODO: Allow more than 10 entries.
        searchRequest.source(sourceBuilder);
        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);
        Collection<String> buckets = new ArrayList<>();
        for (SearchHit searchHit : searchResponse.getHits().getHits()) {
            buckets.add(searchHit.getId());
        }
        return buckets;
    }

    private static void addMetadataQueryShould(BoolQueryBuilder queryBuilder, String value, String field) {
        if (value == null) {
            return;
        }
        queryBuilder.should(QueryBuilders.termsQuery(field, value));
    }
    private static void addMetadataQueryMust(BoolQueryBuilder queryBuilder, String value, String field) {
        if (value == null) {
            return;
        }
        queryBuilder.must(QueryBuilders.simpleQueryStringQuery(value).analyzeWildcard(true).field(field));
    }
}
